# -*- coding: utf-8 -*-
from django.conf import settings
from django.urls import reverse
from django.db import models, IntegrityError
from django.utils.encoding import smart_str
from django.utils.feedgenerator import rfc2822_date

from bbcode.models import BBCodeField
from attachments.models import Attachment
from portailuser.models import PortailUser, Ecole
from portailuser.app_logger import AppLogger
from ldap_inp import LdapClub, LdapAlias, LdapEleve
from ldap_inp.ldapobject import LdapObject
from gestion import fields

from datetime import date, timedelta, datetime
import unicodedata
import os
import re

from unidecode import unidecode

# Réutilise le même logger que gestion.clubs
Logger = AppLogger("gestion_clubs")

LISTE_STAFF = ["president", "vicePresident", "tresorier", "secretaire"]
LISTE_AUTRE = ["displayName", "activite", "mailAlias", "local", "typeClub"]


class ClubManager(models.Manager):
    def par_ident(self, ecole, ident):
        ident_ext = ident + "-" + ecole.ident

        try:
            club_ldap = LdapClub.objects.get(cn=ident_ext, ecole=ecole)
        except LdapClub.DoesNotExist:
            raise Club.DoesNotExist("Club « {} » non trouvé dans le LDAP.".format(ident_ext))

        try:
            club = Club.objects.get(ecole=ecole, ident=ident)

        except Club.DoesNotExist:
            club = Club(ecole=ecole, ident=ident)

        if club.nom != club_ldap.displayName:
            club.nom = club_ldap.displayName
            club.save()

        return club


class Inutilisateur(object):
    def __init__(self, username):
        if not isinstance(username, str):
            username = username
        self.username = username
        self.first_name = "???"
        self.last_name = "????"

    def get_link(self):
        return "<{}> (inexistant)".format(self.username)

    def __str__(self):
        return self.username


def upload_logo_club(instance, filename):
    path = "logos-clubs/"
    name = "{}-{}.png".format(instance.ident, instance.ecole.ident)

    return os.path.join(path, name)


class Club(models.Model):
    ecole = models.ForeignKey(Ecole, editable=False, on_delete=models.CASCADE)
    ident = models.SlugField(max_length=32, editable=False)

    mini_description = models.CharField(max_length=40, blank=True, default="")
    description = BBCodeField(pack="message")

    nom = models.CharField(max_length=64, editable=False)

    blog_act = models.BooleanField(default=False)

    logo = models.ImageField(upload_to=upload_logo_club, blank=True, null=True)

    objects = ClubManager()

    def save(self, *args, **kwargs):
        try:
            # supprime l'ancien logo
            this = Club.objects.get(id=self.id)
            logo_name = "logos-clubs/{}-{}.png".format(self.ident, self.ecole.ident)
            if this.logo == logo_name and not self.logo._committed:
                os.remove(this.logo.path)
                this.logo.delete(save=False)
        except Exception:
            pass

        super(Club, self).save(*args, **kwargs)

    def get_ldap_club(self):
        return LdapClub.objects.get(cn=self.ident_groupe, ecole=self.ecole)

    def get_infos(self):
        club_ldap = self.get_ldap_club()

        infos = {}

        for info in LISTE_STAFF + LISTE_AUTRE:
            valeur = getattr(club_ldap, info)
            if isinstance(valeur, list):
                valeur = list(map(smart_str, valeur))
            elif valeur is not None:
                valeur = smart_str(valeur)
            infos[info] = valeur

        return infos

    # TODO: protéger contre les mises à jour partielles (cette fonction n'est
    # pas "atomique": si 'president' n'est pas la première valeur retournée par
    # valeurs.items(), et que celui-ci est invalide (0 ou >1 valeurs par exemple,
    # on obtient alors une mise à jour partielle !)
    def maj_infos(self, valeurs):
        club_ldap = self.get_ldap_club()
        est_passation = False
        diff = {}

        for (cle, valeur) in valeurs.items():
            if cle not in LISTE_STAFF:
                # Mise à jour des atributs simples
                setattr(club_ldap, cle, valeur)
                continue

            # 'membres' est la liste des nouveaux membres occupant le poste 'cle'
            membres = valeur
            # Liste des anciens membres occupant ce poste
            anciens_responsables = getattr(club_ldap, cle)
            # On se garde du cas où il n'y a qu'un seul membre
            if not isinstance(membres, list):
                membres = [membres]
            if not isinstance(anciens_responsables, list):
                anciens_responsables = [anciens_responsables]

            # Ne rien faire si le/les utilisateurs n'ont pas changés
            if anciens_responsables == [i for i in membres]:
                continue

            if cle == "president":
                # Le changement de président est considéré comme une passation
                est_passation = True

                # Il ne peut y avoir qu'un seul président
                if len(membres) != 1:
                    return None

            diff[cle] = (anciens_responsables, membres)

            # ajoute les personnes non présentes dans le club
            for membre in membres:
                if membre not in club_ldap.memberUid:
                    club_ldap.memberUid.append(membre)

            # Mise à jour du poste
            setattr(club_ldap, cle, valeur)

        if est_passation:
            club_ldap.anneePassation = date.today().year

        club_ldap.save()

        return est_passation, diff

    def ajouter_membres(self, liste):
        club_ldap = self.get_ldap_club()

        member_set = set(club_ldap.memberUid)

        for uid in liste:
            try:
                PortailUser.objects.get_from_ldap(uid)
            except PortailUser.DoesNotExist:
                pass
            else:
                member_set.add(uid)
                # Ajouter des aliases virtuels à l'utilisateur
                ajouter_domaines_membre(club_ldap, uid)

        club_ldap.memberUid = list(member_set)
        club_ldap.save()

    def supprimer_membres(self, liste):
        club_ldap = self.get_ldap_club()

        member_set = set(club_ldap.memberUid)

        for uid in liste:
            member_set.discard(uid)
            # Suppression des aliases virtuels de l'utilisateur
            supprimer_domaines_membre(club_ldap, uid)

        # Les membres du bureau ne peuvent être supprimés
        for poste in LISTE_STAFF:
            membres = getattr(club_ldap, poste)
            if not isinstance(membres, list):
                membres = [membres]

            for membre in membres:
                member_set.add(membre)

        club_ldap.memberUid = list(member_set)
        club_ldap.save()

    def get_membres(self, infos):
        club_ldap = self.get_ldap_club()

        membres = []
        for uid in club_ldap.memberUid:
            try:
                membres.append(PortailUser.objects.get_from_ldap(uid))
            except PortailUser.DoesNotExist:
                membres.append(Inutilisateur(uid))

        membres.sort(key=(lambda x: x.username))
        return membres

    def est_membre(self, membre):
        if hasattr(membre, "username"):
            membre = membre.username

        return bool(LdapClub.objects.filter(cn=self.ident_groupe, ecole=self.ecole, memberUid=membre))

    def get_hote(self):
        return "http://" + settings.URLS_SITES[self.ecole.id - 1]

    @property
    def ident_groupe(self):
        return self.ident + "-" + self.ecole.ident

    def get_absolute_url(self):
        return self.ecole.get_absolute_url() + reverse("club:fiche", args=[self.ident])

    def get_blog_url(self):
        if self.blog_act:
            return self.ecole.get_absolute_url() + reverse("club:liste_articles", args=[self.ident])

        return None

    def __str__(self):
        return self.ident_groupe + ": " + self.nom


def get_user_mail_mapping():
    """
    Retourne un dictionnaire 'utilisateur'->['mailLocal'].
    """

    return {eleve.uid: eleve.mailLocal for eleve in LdapEleve.objects.filter(mailLocal__matching="*")}


def get_emails_for_user(ldap_user, domaine):
    """
    Génère la liste des emails associés à un utilisateur donné, avec le domaine d'alias passé en paramètre.
    """

    user = unidecode(unicodedata.normalize("NFC", f"{ldap_user.prenom}.{ldap_user.nom}")).lower().replace(" ", ".")
    # https://www.jochentopf.com/email/chars.html liste de manière très pratique les caractères
    # à bannir. Par simplicité, je préfère faire une allowlist
    cleaned_user = "".join(c for c in user if re.match("[a-z0-9._]", c) is not None)

    return [f"{cleaned_user}@{domaine}", f"{ldap_user.uid}@{domaine}"]


def update_domaines(ldap_club, to_delete=None):
    """
    Mets à jour les alias virtuels des gens inscrits dans le club. Si `to_delete` n'est pas à None,
    supprime le club associé. `mail_users` est un dictionnaire 'utilisateur'->['mailLocal'].
    """

    # Pas de domaines -> pas de calculs à faire
    if len(ldap_club.domainesAssocies) == 0:
        return

    mail_users = get_user_mail_mapping()

    if to_delete is not None:
        for user, user_mails in mail_users.items():
            ldap_user = LdapEleve.objects.get(uid=user)
            ldap_user_mails = get_emails_for_user(ldap_user, to_delete)

            for mail in user_mails:
                if mail in ldap_user_mails:
                    # le mail a été ajouté automatiquement (l'ajout manuel n'est pas possible,
                    # sauf directement via le LDAP), on peut donc le supprimer
                    ldap_user.mailLocal.remove(mail)
                    ldap_user.save()

        if to_delete in ldap_club.domainesAssocies:
            ldap_club.domainesAssocies.remove(to_delete)
        ldap_club.save()

    # Ajout des alias automatiquement
    for membre in ldap_club.memberUid:
        ajouter_domaines_membre(ldap_club, membre)


def ajouter_domaines_membre(ldap_club, uid):
    """
    Ajouter tous les alias virtuels associé du club auquel l'utilisateur a été inscrit.
    """

    try:
        ldap_user = LdapEleve.objects.get(uid=uid)
    except LdapObject.DoesNotExist:
        return

    source_mails = set(ldap_user.mailLocal)
    mails = source_mails.copy()
    for domaine in ldap_club.domainesAssocies:
        mails |= set(get_emails_for_user(ldap_user, domaine))
    if source_mails != mails:
        ldap_user.mailLocal = list(mails)
        ldap_user.save()


def supprimer_domaines_membre(ldap_club, uid):
    """
    Supprime tous les alias virtuels associé au club auquel l'utilisateur a été désinscrit.
    """

    try:
        ldap_user = LdapEleve.objects.get(uid=uid)
    except LdapObject.DoesNotExist:
        return

    source_mails = set(ldap_user.mailLocal)
    mails = source_mails.copy()
    for domaine in ldap_club.domainesAssocies:
        mails -= set(get_emails_for_user(ldap_user, domaine))
    if source_mails != mails:
        ldap_user.mailLocal = mails
        ldap_user.save()


class ArticleBlog(models.Model):
    club = models.ForeignKey(Club, editable=False, on_delete=models.CASCADE)
    auteur = models.ForeignKey(PortailUser, editable=False, on_delete=models.CASCADE)
    date_post = fields.DateTimeField(auto_now_add=True, editable=False)
    editeur = models.ForeignKey(
        PortailUser,
        editable=False,
        null=True,
        related_name="articleblog_editeur_set",
        on_delete=models.CASCADE,
    )
    date_edit = fields.DateTimeField(auto_now=True, editable=False)

    titre = models.CharField(max_length=64)
    adresse = models.SlugField(max_length=64, editable=False)

    contenu = BBCodeField(pack="message_attach")
    image_accueil = models.ForeignKey(Attachment, null=True, blank=True, editable=False, on_delete=models.SET_NULL)

    sur_accueil = models.BooleanField(editable=False, default=False)
    demande_accueil = models.BooleanField(editable=False, default=False)

    def save(self, *args, **kwargs):
        if not self.adresse:
            raise IntegrityError("Champ adresse vide")
        super(ArticleBlog, self).save(*args, **kwargs)

    def date_r(self):
        return rfc2822_date(self.date_post)

    def get_absolute_url(self):
        return reverse("club:article", args=[self.club.ident, self.adresse])

    def url_image_accueil(self):
        return self.image_accueil.get_mini_url() if self.image_accueil else None

    class Meta:
        unique_together = (("club", "adresse"),)


def date_limite_edition():
    return datetime.now() + timedelta(minutes=settings.EDITION_TIME)


class JetonEdition(models.Model):
    article = models.OneToOneField(ArticleBlog, primary_key=True, on_delete=models.CASCADE)
    editeur = models.ForeignKey(PortailUser, on_delete=models.CASCADE)
    fin_jeton = models.DateTimeField(default=date_limite_edition)

    def renouvellement(self):
        self.fin_jeton = date_limite_edition()
        self.save()


class AliasGere(models.Model):
    club = models.ForeignKey(Club, editable=False, on_delete=models.CASCADE)
    prefixe = models.CharField(max_length=32, verbose_name="Préfixe")
    confirme = models.BooleanField(default=False, editable=False)

    def __str__(self):
        return self.prefixe

    @property
    def adresse(self):
        return "{}@{}".format(self.prefixe, settings.EMAIL_BDE[self.club.ecole.id - 1])

    @property
    def alias_ldap(self):
        return LdapAlias.objects.get(cn=self.prefixe, base=self.club.ecole.dn)

    class Meta:
        unique_together = (("club", "prefixe"),)


class MesBlogs(models.Model):
    """Widget blogs"""

    club_fav = models.ForeignKey(Club, on_delete=models.CASCADE)
    utilisateur = models.ForeignKey(PortailUser, on_delete=models.CASCADE)
    ecole = models.ForeignKey(Ecole, on_delete=models.CASCADE)
